/* ==========================================================================*/
/* - - - - - - - - - - - - - -  I r q . c  - - - - - - - - - - - - - - - - - */
/* ==========================================================================*/
/*	Created   2011-05-29   M. Wheatley                                       */
/*	Modified  2012-04-08   MW  Updated FreeRTOS to 7.1.0                     */
/*.........................................................................  */
/* WSPRbeacon  IRQ service Routines                                          */
/*     (must be compiled as Non-Thum ARM mode)                               */
/*.........................................................................  */
///////////////////////////////////////////////////////////////////////////////
//            2011 Copyright (C) Moe Wheatley, AE4JY
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//////////////////////////////////////////////////////////////////////////////

/*---------------------------------------------------------------------------*/
/*------------------------> I N C L U D E S <--------------------------------*/
/*---------------------------------------------------------------------------*/
#include "common.h"
#include "board.h"
#include "usart.h"
#include "timer.h"
#include "irq.h"
#include "gps.h"
#include "term.h"
#include "wspr.h"


/* Scheduler includes. */
#include "FreeRTOS.h"
#include "semphr.h"
#include "task.h"

/*---------------------------------------------------------------------------*/
/*--------------------> L O C A L   D E F I N E S <--------------------------*/
/*---------------------------------------------------------------------------*/
#define PPS_TIMEOUT 500

/*---------------------------------------------------------------------------*/
/*---------------------> L O C A L   M E M O R Y  <--------------------------*/
/*---------------------------------------------------------------------------*/
volatile BOOL MsgAbort;
volatile uint32_t MsgTimer;


/*---------------------------------------------------------------------------*/
/*------------------------> M A C R O S <------------------------------------*/
/*---------------------------------------------------------------------------*/


	 	
	 	
/******************************************************************************
 *
 * MACRO Name: ISR_ENTRY()
 *
 * Description:
 *    This MACRO is used upon entry to an ISR.  The current version of
 *    the gcc compiler for ARM does not produce correct code for
 *    interrupt routines to operate properly with THUMB code.  The MACRO
 *    performs the following steps:
 *
 *    1 - Adjust address at which execution should resume after servicing
 *        ISR to compensate for IRQ entry
 *    2 - Save the non-banked registers r0-r12 and lr onto the IRQ stack.
 *    3 - Get the status of the interrupted program is in SPSR.
 *    4 - Push it onto the IRQ stack as well.
 *
 *****************************************************************************/
#define ISR_ENTRY() asm volatile(" sub   lr, lr,#4\n" \
                                 " stmfd sp!,{r0-r12,lr}\n" \
                                 " mrs   r1, spsr\n" \
                                 " stmfd sp!,{r1}")
/******************************************************************************
 *
 * MACRO Name: ISR_EXIT()
 *
 * Description:
 *    This MACRO is used to exit an ISR.  The current version of the gcc
 *    compiler for ARM does not produce correct code for interrupt
 *    routines to operate properly with THUMB code.  The MACRO performs
 *    the following steps:
 *
 *    1 - Recover SPSR value from stack       
 *    2 - and restore  its value                   
 *    3 - Pop the return address & the saved general registers from
 *        the IRQ stack & return
 *
 *****************************************************************************/
#define ISR_EXIT()  asm volatile(" ldmfd sp!,{r1}\n" \
                                 " msr   spsr_c,r1\n" \
                                 " ldmfd sp!,{r0-r12,pc}^") 
  

/******************************************************************************
 *
 * MACRO Name: FSR_ENTRY()
 *
 * Description:
 *    This MACRO is used upon entry to an FSR.  The current version of
 *    the gcc compiler for ARM does not produce correct code for
 *    interrupt routines to operate properly with THUMB code.  The MACRO
 *    performs the following steps:
 *
 *    1 - Adjust address at which execution should resume after servicing
 *        FSR to compensate for FIQ entry
 *    2 - Save the non-banked registers r0-r7 and lr onto the FIQ stack.
 *
 *****************************************************************************/
#define FSR_ENTRY() asm volatile(" sub   lr, lr,#4\n" \
                                 " stmfd sp!,{r0-r7,lr}")

/******************************************************************************
 *
 * MACRO Name: FSR_EXIT()
 *
 * Description:
 *    This MACRO is used to exit an FSR.  The current version of the gcc
 *    compiler for ARM does not produce correct code for interrupt
 *    routines to operate properly with THUMB code.  The MACRO performs
 *    the following steps:
 *
 *    1 - Pop the return address & the saved general registers from
 *        the FIQ stack & return
 *
 *****************************************************************************/
#define FSR_EXIT()  asm volatile(" ldmfd sp!,{r0-r7,pc}^")


void vUsart0IrqService(void);
void vGpsIrqService(void);
void vTimer2IrqService(void);
void vTimer1IrqService(void);
void vTimer0IrqService(void);

//wrappers for irq routines


void Usart0IrqService(void)
{
	portSAVE_CONTEXT();
	vUsart0IrqService();
	portRESTORE_CONTEXT();
}
void GpsIrqService(void)
{
	portSAVE_CONTEXT();
	vGpsIrqService();
	portRESTORE_CONTEXT();
}

void Timer2IrqService(void)
{
	portSAVE_CONTEXT();
	vTimer2IrqService();
	portRESTORE_CONTEXT();
}
void Timer1IrqService(void)
{
	portSAVE_CONTEXT();
	vTimer1IrqService();
	portRESTORE_CONTEXT();
}
void Timer0IrqService(void)
{
	portSAVE_CONTEXT();
	vTimer0IrqService();
	portRESTORE_CONTEXT();
}

/*---------------------------------------------------------------------------*/
/*----------> I N T E R R U P T    S E R V I C E    R O U T I N E S <--------*/
/*---------------------------------------------------------------------------*/

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                    ---------------------------------                      */
/*                   | U s a r t 0 I r q S e r v i c e |                     */
/*                    ---------------------------------                      */
/*    Called by usart0 IRQ.                                                  */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void vUsart0IrqService(void)
{
unsigned int data __attribute__((unused));
portBASE_TYPE TaskWoken = pdFALSE;
	AT91C_BASE_AIC->AIC_EOICR = 0;
#if REV1_PCB	//usart0 used for debug
	uint32_t Status = *AT91C_US0_CSR;
	if ( Status & AT91C_US_RXRDY)
	{
		if ( 0 == (Status & (AT91C_US_OVRE|AT91C_US_FRAME)) )
		{	//* Get byte
			UsartRxQueue[UsartRxHead++] = (uint8_t)*AT91C_US0_RHR;
			UsartRxHead &= (USARTQSIZE-1);	//deal with rollover
			xSemaphoreGiveFromISR( xUsartRxSemaphore, &TaskWoken );
		}
		else
		{
			data = *AT91C_US0_RHR;
		}
	}
	if( Status & AT91C_US_TXRDY)
	{
		if( UsartTxHead != UsartTxTail )
		{
			*AT91C_US0_THR = UsartTxQueue[UsartTxTail++];
			UsartTxTail &= (USARTQSIZE-1);	//deal with rollover
		}
		else
		{
			*AT91C_US0_IDR = AT91C_US_TXRDY;
		}
	}
	//* Reset the status bit
	*AT91C_US0_CR = AT91C_US_RSTSTA;
#else	//USART0 used for terminal
	uint32_t Status = *AT91C_US0_CSR;
	if ( Status & AT91C_US_RXRDY)
	{
		if ( 0 == (Status & (AT91C_US_OVRE|AT91C_US_FRAME)) )
		{	//* Get byte
			TermRxQueue[TermRxHead++] = (uint8_t)*AT91C_US0_RHR;
			TermRxHead &= (TERMQSIZE-1);	//deal with rollover
			xSemaphoreGiveFromISR( xTermRxSemaphore, &TaskWoken );
		}
		else
		{
			data = *AT91C_US0_RHR;
		}
	}
	if( Status & AT91C_US_TXRDY)
	{
		if( TermTxHead != TermTxTail )
		{
			*AT91C_US0_THR = TermTxQueue[TermTxTail++];
			TermTxTail &= (TERMQSIZE-1);	//deal with rollover
		}
		else
		{
			*AT91C_US0_IDR = AT91C_US_TXRDY;
		}
	}
	//* Reset the status bit
	*AT91C_US0_CR = AT91C_US_RSTSTA;
#endif
	if( TaskWoken )
		portYIELD_FROM_ISR();
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                    ---------------------------                            */
/*                   | G p s I r q S e r v i c e |                           */
/*                    ---------------------------                            */
/*    Called by GPS usart1 IRQ.                                              */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void vGpsIrqService(void)
{
uint32_t status;
portBASE_TYPE TaskWoken = pdFALSE;
	AT91C_BASE_AIC->AIC_EOICR = 0;
	status = *AT91C_US1_CSR;
	if(status & AT91C_US_RXRDY )
	{
		if( !(status & (AT91C_US_OVRE|AT91C_US_FRAME)) )
		{
			GpsRxQueue[GpsRxHead++] = *AT91C_US1_RHR;
			if(GpsRxHead>=GPSQSIZE)
				GpsRxHead = 0;
			 xSemaphoreGiveFromISR( xGpsRxSemaphore, &TaskWoken);
		}
	}
	AT91C_BASE_US1->US_CR = AT91C_US_RSTSTA;
	if( TaskWoken )
		portYIELD_FROM_ISR();
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                    -----------------------------                          */
/*                   | D B G U I r q S e r v i c e |                         */
/*                    -----------------------------                          */
/*    Called by DBGU usart IRQ.                                              */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void DBGUIrqService(void)
{
volatile unsigned int data __attribute__((unused));
portBASE_TYPE TaskWoken = pdFALSE;
#if REV1_PCB	//DBGU uart used for terminal
uint32_t Status = *AT91C_DBGU_CSR;
	if ( Status & AT91C_US_RXRDY)
	{
		if ( 0 == (Status & (AT91C_US_OVRE|AT91C_US_FRAME)) )
		{	//* Get byte
			TermRxQueue[TermRxHead++] = (uint8_t)*AT91C_DBGU_RHR;
			TermRxHead &= (TERMQSIZE-1);	//deal with rollover
			xSemaphoreGiveFromISR( xTermRxSemaphore, &TaskWoken);
		}
		else
		{
			data = *AT91C_DBGU_RHR;
		}
	}
	if( Status & AT91C_US_TXRDY)
	{
		if( TermTxHead != TermTxTail )
		{
			*AT91C_DBGU_THR = TermTxQueue[TermTxTail++];
			TermTxTail &= (TERMQSIZE-1);	//deal with rollover
		}
		else
		{
			*AT91C_DBGU_IDR = AT91C_US_TXRDY;
		}
	}
	//* Reset the status bit
	*AT91C_DBGU_CR = AT91C_US_RSTSTA;
#else	//DBGU uart used for debug

	uint32_t Status = *AT91C_DBGU_CSR;
	if ( Status & AT91C_US_RXRDY)
	{
		if ( 0 == (Status & (AT91C_US_OVRE|AT91C_US_FRAME)) )
		{	//* Get byte
			UsartRxQueue[UsartRxHead++] = (uint8_t)*AT91C_DBGU_RHR;
			UsartRxHead &= (USARTQSIZE-1);	//deal with rollover
			xSemaphoreGiveFromISR( xUsartRxSemaphore, &TaskWoken );
		}
		else
		{
			data = *AT91C_DBGU_RHR;
		}
	}
	if( Status & AT91C_US_TXRDY)
	{
		if( UsartTxHead != UsartTxTail )
		{
			*AT91C_DBGU_THR = UsartTxQueue[UsartTxTail++];
			UsartTxTail &= (USARTQSIZE-1);	//deal with rollover
		}
		else
		{
			*AT91C_DBGU_IDR = AT91C_US_TXRDY;
		}
	}
	//* Reset the status bit
	*AT91C_DBGU_CR = AT91C_US_RSTSTA;
#endif
	if( TaskWoken )
		portYIELD_FROM_ISR();
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      ---------------------------------                    */
/*                     | T i m e r 0 I r q S e r v i c e |                   */
/*                      ---------------------------------                    */
/*    Called by IRQ to service timer 0 at 1.46Hz WSPR symbol time.           */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void vTimer0IrqService(void)
{
volatile uint32_t status __attribute__((unused));
portBASE_TYPE TaskWoken = pdFALSE;
uint8_t Cmd;
    //* Acknowledge interrupt status
	status = AT91C_BASE_TC0->TC_SR;
	AT91C_BASE_AIC->AIC_EOICR = 0;
	if(SymbCount != 0)
	{
		Cmd = WPSRCMD_BIT;
		xQueueSendToBackFromISR( xWsprCmdQueue, &Cmd, &TaskWoken);
	}
	if( TaskWoken )
		portYIELD_FROM_ISR();
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      ---------------------------------                    */
/*                     | T i m e r 1 I r q S e r v i c e |                   */
/*                      ---------------------------------                    */
/*    Capture Timer for 1pps GPS signal.  timer clk is 24 MHz                */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void vTimer1IrqService(void)
{
volatile uint32_t status __attribute__((unused));
	//* Acknowledge interrupt status
	AT91C_BASE_AIC->AIC_EOICR = 0;
	status = AT91C_BASE_TC1->TC_SR;
	if(status&AT91C_TC_COVFS)
	{	// 16 bit counter overflow so just inc msb
		CaptureTime += 0x00010000;
		if(PpsTimeout)
		{
			PpsTimeout--;
			if(PpsTimeout<=0)
				PpsPeriod = 0;
		}
	}
	//in free run capture mode, irq toggles between capture A and capture B so service both
	if(status&AT91C_TC_LDRAS)
	{	// 1PPS positive edge occured so get capture time
		CaptureTime = (CaptureTime&0xFFFF0000) + AT91C_BASE_TC1->TC_RA;
		PpsPeriod = CaptureTime - LastCaptureTime;
		LastCaptureTime = CaptureTime;
		PpsTimeout = PPS_TIMEOUT;	//reset 1PPS timeout timer
		//sync 100Hz timer2 to the 1PPS edge
		AT91C_BASE_TC2->TC_CCR = AT91C_TC_SWTRG;	//reset timer
		if(Timer100Hz>50)
		{
			Timer100Hz = 100;
			AT91C_BASE_TC2->TC_IER = AT91C_TC_CPAS;  //  IRQ enable
		}
		else
		{
			Timer100Hz = 0;
		}
	}
	if(status&AT91C_TC_LDRBS)
	{	// 1PPS positive edge occured so get capture time
		CaptureTime = (CaptureTime&0xFFFF0000) + AT91C_BASE_TC1->TC_RB;
		PpsPeriod = CaptureTime - LastCaptureTime;
		LastCaptureTime = CaptureTime;
		PpsTimeout = PPS_TIMEOUT;	//reset 1PPS timeout timer
		//sync 100Hz timer2 to the 1PPS edge
		AT91C_BASE_TC2->TC_CCR = AT91C_TC_SWTRG;	//reset timer
		if(Timer100Hz>50)
		{
			Timer100Hz = 100;
			AT91C_BASE_TC2->TC_IER = AT91C_TC_CPAS;  //  IRQ enable
		}
		else
		{
			Timer100Hz = 0;
		}
	}
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      ---------------------------------                    */
/*                     | T i m e r 2 I r q S e r v i c e |                   */
/*                      ---------------------------------                    */
/*    Called by IRQ to service timer 2 at 100Hz rate.                        */
/* Locked to GPS 1PPS if available else free runs using TCXO.                */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void vTimer2IrqService(void)
{
volatile uint32_t status __attribute__((unused));
portBASE_TYPE TaskWoken = pdFALSE;
uint8_t Cmd;
    //* Acknowledge interrupt status
	AT91C_BASE_AIC->AIC_EOICR = 0;
	status = AT91C_BASE_TC2->TC_SR;
	if( status&AT91C_TC_CPAS)
		AT91C_BASE_TC2->TC_IDR = AT91C_TC_CPAS;
	if(++Timer100Hz >= 100)	//see if counted to 1 second
	{
		Timer100Hz = 0;
		Cmd = WPSRCMD_SECOND;
		xQueueSendToBackFromISR( xWsprCmdQueue, &Cmd, &TaskWoken);
		if(++SecondTimer >= 60)
		{
			SecondTimer = 0;
			if(++MinuteTimer >= 60)
			{
				MinuteTimer = 0;
				if(++HourTimer >= 60)
				{
					HourTimer = 0;
				}
			}
		}
		if(GpsTimeIsValid)
		{	//if GPS time is valid update local time values
			SecondTimer = GPSSecond;
			MinuteTimer = GPSMinute;
			HourTimer = GPSHour;
		}
	}
//Check for Si5338 PLL error and disable TX amp if error
	if( AT91F_PIO_GetInput(AT91C_BASE_PIOA) & INT5338 )
	{
		if( AT91F_PIO_GetInput(AT91C_BASE_PIOA) & TXOFF )
			AT91F_PIO_SetOutput( AT91C_BASE_PIOA, TXOFF);
	}
	if( TaskWoken )
		portYIELD_FROM_ISR();
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                   -------------------------------------                   */
/*                  | S p u r i o u s I r q S e r v i c e |                  */
/*                   -------------------------------------                   */
/*  Must have do nothing handler for spurious interrupts.                    */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void SpuriousIrqService(void)
{
	ISR_ENTRY();
	AT91C_BASE_AIC->AIC_EOICR = 0;
	ISR_EXIT();
}

