/* ==========================================================================*/
/* - - - - - - - - - - - - -  s i 5 3 3 8 . c  - - - - - - - - - - - - - - - */
/* ==========================================================================*/
/*	Created   2011-09-21   M. Wheatley                                       */
/*	Modified  2011-09-21   MW                                                */
/*.........................................................................  */
/* WSPR Beacon   AT91SAM7256 SI 5338 clock generator management              */
///////////////////////////////////////////////////////////////////////////////
//            2011 Copyright (C) Moe Wheatley, AE4JY
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//////////////////////////////////////////////////////////////////////////////

/*---------------------------------------------------------------------------*/
/*------------------------> I N C L U D E S <--------------------------------*/
/*---------------------------------------------------------------------------*/
#include "board.h"
#include "common.h"
#include "si5338.h"
#include "timer.h"
#include "siregvals.h"
#include "iicctrl.h"
#include "protocoldefs.h"
#include "stdlib.h"

/*---------------------------------------------------------------------------*/
/*--------------------> L O C A L   D E F I N E S <--------------------------*/
/*---------------------------------------------------------------------------*/
#define FREQ_PFD 25000000ull
#define NOMVCO_FREQ 2500000000ull
#define MINVCO_FREQ 2200000000ull
#define MAXVCO_FREQ 2846000000ull

#define LOCK_MASK 0x15
#define LOS_MASK 0x04

/*---------------------------------------------------------------------------*/
/*-------------------> M E M O R Y  A L L O C A T I O N <--------------------*/
/*---------------------------------------------------------------------------*/
BOOL RfOnState;
static int SiPage;
static uint32_t Val_R;	//output divider value
static uint32_t Val_MS0;	//MultiSynth MS0 integer divider value(4 to 568 even only)
static uint32_t Val_A;	//PLL integer divider
static uint32_t Val_B;	//PLL fraction numerator
static uint32_t Val_C;	//PLL fraction denominator
static int32_t CurrentWsprSymb;


/*---------------------------------------------------------------------------*/
/*--------------------------> P R O T O T Y P E S <--------------------------*/
/*---------------------------------------------------------------------------*/
static void WriteSi5338Reg(uint32_t reg, uint8_t val);
static uint64_t GCD(uint64_t u, uint64_t v);
static void CalcPLLabc(uint32_t freq);
static void CalcPLLparameters(uint64_t VcoFreq);

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      ---------------------                                */
/*                     | S i 5 3 3 8 I n i t |                               */
/*                      ---------------------                                */
/*    Called to initialize the Si5338 clock chip.                            */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void Si5338Init ( void )
{
int i;
	WriteSi( 255,0);		//select page 0
	SiPage = 0;

	WriteSi5338Reg(110,0x40);	//output inactive low
	WriteSi5338Reg(230,0x1F);	//disable outputs

	//clear phase adjust regs
	for(i=107; i<=122; i++)
		WriteSi5338Reg(i,0 );

	//clear spread spectrum regs
	for(i=287; i<=347; i++)
		WriteSi5338Reg(i,0 );

	//configure input selections
	WriteSi5338Reg(28,0x8B);	//26 to 30MHz input on single ended IN3
	WriteSi5338Reg(29,0x08);	//single ended IN3, P1DIV=1, PFD_INREF == refclk
	WriteSi5338Reg(30,0xB0);	//P2DIV=1,  PFD_FB == noclk, No lk to P2

	//configure output drivers
	//	signal type CMOS A en B dis  unused outputs==LVDS no inversion
	WriteSi5338Reg(36,0x01);
	WriteSi5338Reg(37,0x06);
	WriteSi5338Reg(38,0x06);
	WriteSi5338Reg(39,0x06);

	//configure output voltage 3.3V
	WriteSi5338Reg(35,0x00);

	//configure output Trim  3.3V CMOS
	WriteSi5338Reg(40,0x77);	//011 10111
	WriteSi5338Reg(41,0x0C);	//0 00011 00
	WriteSi5338Reg(42,0x23);	//001 00011

	//configure output disable states
	WriteSi5338Reg(110,0x40);	//output inative low
	WriteSi5338Reg(114,0x40);	//output inative low
	WriteSi5338Reg(118,0x40);	//output inative low
	WriteSi5338Reg(122,0x40);	//output inative low

	//configure output clock source select
	WriteSi5338Reg(31,0xC0);	//clk0 uses Multisynth 0
	WriteSi5338Reg(32,0xE3);	//not used
	WriteSi5338Reg(33,0xE3);	//not used
	WriteSi5338Reg(34,0xE3);	//not used

	//Misc register setup as per spec
	WriteSi5338Reg(47,0x14);	//set value if not factory programmed
	WriteSi5338Reg(6,0x08);		//INT is fb lock detect
	WriteSi5338Reg(116,0x80);	//set bit 7 per spec

	RfOnState = true;
	RfOff();
	SetFreq( FREQBASE_TBL[Settings.BandIndex] + Settings.FreqOffset[Settings.BandIndex]);
}

//////////////////////////////////////////////////////////////////////////////////
//  Sets the Si5338 frequency.
//////////////////////////////////////////////////////////////////////////////////
void SetFreq(uint32_t freq)
{
uint64_t tmp;
uint8_t btmp;
int Timeout;
	RfOff();
	WriteSi5338Reg(241, 0xE5); //DIS_LOL = 1
	CalcPLLabc(freq);
	//set output divider value
	if(1==Val_R) btmp = 0;
	else if(2==Val_R) btmp = 1<<2;
	else if(4==Val_R) btmp = 2<<2;
	else if(8==Val_R) btmp = 3<<2;
	else if(16==Val_R) btmp = 4<<2;
	else btmp = 5<<2;
	WriteSi5338Reg(31,0xC0+btmp);	//clk0 uses Multisynth 0 and Val_R output divider

	//calc MS0_P1,2,3 output multisynth register values
	// MS0_P1 = floor[ ((a*c+b)*128)/c - 512]	//a == Val_MS0, c==1, b==0
	tmp = ( (uint64_t)Val_MS0* 128ul ) - 512ul;
	WriteSi5338Reg(53, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(54, tmp&0xFF);
	tmp >>= 8;
	btmp = tmp & 0x03;	//save bits 16,17 to put in reg 55
	// MS0_P2 = (b*128) % c		//always 0 since c==1
	WriteSi5338Reg(55,  btmp);
	WriteSi5338Reg(56, 0);
	WriteSi5338Reg(57, 0);
	WriteSi5338Reg(58, 0);
	// MS0_P3 = c		//c==1
	WriteSi5338Reg(59, 1);
	WriteSi5338Reg(60, 0);
	WriteSi5338Reg(61, 0);
	WriteSi5338Reg(62, 0);

	//calc MSN_P1,2,3 register values
	// MSN_P1 = floor[ ((a*c+b)*128)/c - 512]
	tmp = ( ((uint64_t)Val_A*(uint64_t)Val_C+(uint64_t)Val_B) * 128ull) / (uint64_t)Val_C - 512ull;
//	printf("P1 = %d\r\n",(uint32_t)tmp);
	WriteSi5338Reg(97, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(98, tmp&0xFF);
	tmp >>= 8;
	btmp = tmp & 0x03;	//save bits 16,17 to put in reg 99
	// MSN_P2 = (b*128) % c
	tmp = ( (uint64_t)Val_B * 128ull) % (uint64_t)Val_C;
//	printf("P2 = %d\r\n",(uint32_t)tmp);
	tmp <<= 2;
	WriteSi5338Reg(99, (tmp&0xFF) + btmp);
	tmp >>= 8;
	WriteSi5338Reg(100, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(101, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(102, tmp&0xFF);
	// MSN_P3 = c
	tmp = (uint64_t)Val_C;
//	printf("P3 = %d\r\n",Val_C);
	WriteSi5338Reg(103, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(104, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(105, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(106, 0x80 | (tmp&0xFF));	//bit 7 must be set as per spec

	//set inc step size to WSPR step = 375/256 Hz = 1.46484375Hz  = 12000/8192
	//FIDP1 = c * freq * 256/375  where c==1
	tmp = ( (uint64_t)freq * 256ull)/375ull;
	WriteSi5338Reg(123, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(124, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(125, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(126, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(127, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(128, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(129, tmp&0xFF);
	//FIDP2 = c = 1
	WriteSi5338Reg(136, 1);
	WriteSi5338Reg(135, 0);
	WriteSi5338Reg(134, 0);
	WriteSi5338Reg(133, 0);
	WriteSi5338Reg(132, 0);
	WriteSi5338Reg(131, 0);
	WriteSi5338Reg(130, 0);
	//FIDP3 = (a*c+b)*Freq/Fstep == Val_MS0 * freq * 256/375
	tmp = ( (uint64_t)Val_MS0*(uint64_t)freq * 256ull)/375ull;
	WriteSi5338Reg(137, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(138, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(139, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(140, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(141, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(142, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(143, tmp&0xFF);
	tmp >>= 8;
	WriteSi5338Reg(144, tmp&0xFF);

	Timeout = 200;
	do
	{
		btmp = ReadSi(218) & LOS_MASK;
		vTaskDelay(1);
		Timeout--;
	}
	while(Timeout && (btmp != 0) );

	WriteSi5338Reg(49, ReadSi(49) & 0x7F); //FCAL_OVRD_EN = 0
	WriteSi5338Reg(246, 2); //soft reset
	WriteSi5338Reg(241, 0x65); //DIS_LOL = 0

	vTaskDelay(50);
	Timeout = 200;
	do
	{
		btmp = ReadSi(218) & LOCK_MASK;
		Timeout--;
	}
	while(Timeout && (btmp != 0) );

	//copy FCAL values
	WriteSi5338Reg(45, ReadSi(235));
	WriteSi5338Reg(46, ReadSi(236));
	// clear bits 0 and 1 from 47 and
	// combine with bits 0 and 1 from 237
	btmp = (ReadSi(47) & 0xFC) | (ReadSi(237) & 3);
	WriteSi5338Reg(47, btmp);
	WriteSi5338Reg(49, ReadSi(49) | 0x80); // FCAL_OVRD_EN = 1
}

//////////////////////////////////////////////////////////////////////////////////
//  Initializes Si5338 frequency and symbol state and enables PA for transmit
//////////////////////////////////////////////////////////////////////////////////
void RfOn()
{
	if(!RfOnState)
	{
		WriteSi5338Reg(52, 0x10);	//set back to base frequency
		WriteSi5338Reg(52, 0x00);	//
		CurrentWsprSymb = 0;
		WriteSi5338Reg(230, 0x0E);
		if( 0 == (AT91F_PIO_GetInput(AT91C_BASE_PIOA) & INT5338) )
		{
			RfOnState = true;
			AT91F_PIO_ClearOutput( AT91C_BASE_PIOA, TXOFF);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////
//  Turns off RF and disables PA
//////////////////////////////////////////////////////////////////////////////////
void RfOff()
{
	AT91F_PIO_SetOutput( AT91C_BASE_PIOA, TXOFF);
	if(RfOnState)
	{
		WriteSi5338Reg(230, 0x1F);
		RfOnState = false;
	}

}

//////////////////////////////////////////////////////////////////////////////////
//  Sets WSPR symbol frequency using INC DEC commands to Si5338 for smooth changes
//////////////////////////////////////////////////////////////////////////////////
void SetWsprSymbFreq(int32_t symb)
{
int32_t delta;
	delta = symb - CurrentWsprSymb;
	if(0 == delta)
		return;
	if(delta > 0)	//need to inc
	{
		while(delta--)
		{
			WriteSi5338Reg(52, 0x40);	//inc
			CurrentWsprSymb++;
		}
	}
	else
	{
		while(delta++)
		{
			WriteSi5338Reg(52, 0x60);	//dec
			CurrentWsprSymb--;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////
//calculate PLL gain and BW parameters from vco freq
//////////////////////////////////////////////////////////////////////////////////
static void CalcPLLparameters(uint64_t VcoFreq)
{
uint64_t tmp;
uint32_t K;
uint32_t Q;
uint8_t GN;
uint64_t x;
	if(VcoFreq>2425000000ull)
	{
		Q = 3;
		GN = 0<<4;
	}
	else
	{
		Q = 4;
		GN = 1<<4;
	}
	if(FREQ_PFD>=15000000ull)
	{
		K = 925;
		GN |= 0;	//add RSEL and BWSEL bits
	}
	else if( (FREQ_PFD>=8000000ull) && (FREQ_PFD<15000000ull) )
	{
		K = 325;
		GN |= 0x05;	//add RSEL and BWSEL bits
	}
	else
	{
		K = 185;
		GN |= 0x0E;	//add RSEL and BWSEL bits
	}
	WriteSi5338Reg(49, GN);	//write gain and SEL bits

	//calc MSCAL value
	// MSCAL = ROUND( -6.67*VcoFreq/1e9 + 20.67)
	//use 64 bit integer math and scaling
	tmp = ( (2067ull - ( (667ull*VcoFreq)/1000000000ull) ) + 50ull) /100ull;
	WriteSi5338Reg(50, 0xC0 + (tmp&0x3F));	//write MSCAL value

	//calc PLL_KPH value                      |         | 3
	//                    K        Fvco       | Fvconom |
	//PLL_KPH = ROUND[ ------- * ---------- * |-------- |
	//                  533*Q      Fpfd       |  Fvco   |
	///////////////////////////////////////////////////////
	x = (NOMVCO_FREQ*1000ull)/VcoFreq;
	tmp = ( (uint64_t)K*VcoFreq)/FREQ_PFD;
	tmp  = tmp*x*x*x + 1000000000000ull;
	tmp = tmp/(533ull * Q * 1000000000ull);
	WriteSi5338Reg(48, tmp&0x7F);	//write PLL_KPH value
	//set MS_PEC to 0x07 as per spec
	WriteSi5338Reg( 51, 0x07);
}

//////////////////////////////////////////////////////////////////////////////////
//calculate PLL registers a,b, and c as well as output MS0 divider values N and R
//////////////////////////////////////////////////////////////////////////////////
static void CalcPLLabc(uint32_t freq)
{
uint32_t m;
uint64_t minerror;
uint64_t VcoFreq;
uint64_t error;
	//calc output integer value
	Val_R = 1;
	while( (Val_R <= 32) && (Val_R*freq)< 5000000)
		Val_R <<= 1;
	//calc output multisyth divider value. Force to be an integer from 4 to 568 even only
	Val_MS0 = 1;
	minerror = MAXVCO_FREQ;
	for(m=4; m<=568; m+=2)
	{
		VcoFreq = (uint64_t)freq * (uint64_t)m * (uint64_t)Val_R;
		//only look at values that keep Fvco in range
		if( (VcoFreq>MINVCO_FREQ) && (VcoFreq<MAXVCO_FREQ) )
		{
			//keep value that is closest to nominal Fvco value
			if(VcoFreq>NOMVCO_FREQ)
				error = VcoFreq - NOMVCO_FREQ;
			else
				error = NOMVCO_FREQ - VcoFreq;
			if(error<minerror)
			{	//save value closest to nominal Fvco
				minerror = error;
				Val_MS0 = m;
			}
		}
	}
	//calc Fvco using the nominal output divider values found above
	VcoFreq = (uint64_t)freq * (uint64_t)Val_MS0 * (uint64_t)Val_R;
	//calc pll feedback fractional divider in form a + b/c
	Val_A = VcoFreq/FREQ_PFD;		//get integer part
	uint32_t rem = VcoFreq%FREQ_PFD;	//calc remainder of fraction
	uint64_t gcd = GCD(rem,FREQ_PFD);	//reduce fraction by greatest common denominator
	Val_B = (uint64_t)rem/gcd;
	Val_C = FREQ_PFD/gcd;
//printf("\r\nFreq = %u  Val_R=%u Val_M=%u\r\n",freq, Val_R, Val_MS0);
//printf("A=%u  B=%u  C=%u  \r\n",Val_A, Val_B, Val_C );
	CalcPLLparameters(VcoFreq);
}

/////////////////////////////////////////////////////////////////////////
//Binary Greatest Common Denominator algorithm or Stein's Algorithm
//returns GCD of u and v.
/////////////////////////////////////////////////////////////////////////
static uint64_t GCD(uint64_t u, uint64_t v)
{
int shift;
	/* GCD(0,x) := x */
	if(u == 0 || v == 0)
		return u | v;

	/* Let shift := lg K, where K is the greatest power of 2
		dividing both u and v. */
	for(shift = 0; ((u | v) & 1) == 0; ++shift)
	{
		u >>= 1;
		v >>= 1;
	}
	while((u & 1) == 0)
		u >>= 1;
	/* From here on, u is always odd. */
	do{
		while ((v & 1) == 0)  /* Loop X */
			v >>= 1;

		/* Now u and v are both odd, so diff(u, v) is even.
		Let u = min(u, v), v = diff(u, v)/2. */
		if (u < v)
		{
			v -= u;
		}
		else
		{
			uint64_t diff = u - v;
			u = v;
			v = diff;
		}
		v >>= 1;
	}while(v != 0);
	return u << shift;
}

////////////////////////////////////////////////////////////////////////
// write val to reg managing the mask values and page state
////////////////////////////////////////////////////////////////////////
static void WriteSi5338Reg(uint32_t reg, uint8_t val)
{
uint8_t radr;
uint8_t tmp;
uint8_t mask;
	if(0 == SIMASK_TBL[reg] )
		return;	//can't write register if mask is zero
	//see what page its in and change page if needed
	if(reg>255)
	{
		if(0==SiPage)
		{
			WriteSi( 255,1);		//select page 1
			SiPage = 1;
		}
	}
	else
	{
		if(1==SiPage)
		{
			WriteSi( 255,0);		//select page 0
			SiPage = 0;
		}
	}
	radr = reg & 0x00FF;
	mask = SIMASK_TBL[reg];
	if(0xFF == mask )
	{	//if can write all bits then just write value
		WriteSi( radr, val );
	}
	else	//must do read modify write if mask has partial bits set
	{
		tmp = ReadSi(radr);	//read current value
		tmp = tmp & ~mask;	//clear all non masked bits
		val &= mask;		//get only masked new bits
		val |= tmp;			//OR them with the previous register bits
		WriteSi( radr, val );
	}
}
